#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Semversioner allows you to manage semantic versioning properly and simplifies changelog generation. 

This project was inspired by the way AWS manages their versioning for AWS-CLI: https://github.com/aws/aws-cli/

At any given time, the ``.semversioner/`` directory looks like:
    .semversioner
    |
    └── next-release
        ├── minor-20181227010225.json
        └── major-20181228010225.json
    ├── 1.1.0.json
    ├── 1.1.1.json
    ├── 1.1.2.json

This script takes everything in ``next-release`` and aggregates them all together in a single JSON file for that release (e.g ``1.12.0.json``).  This
JSON file is a list of all the individual JSON files from ``next-release``.

This is done to simplify changelog generation.

Usage
=====
::
    $ semversioner add-change --type major --description "This description will appear in the change log"
    $ semversioner release
    $ semversioner changelog > CHANGELOG.md
"""

import os
import sys
import json
import click
import datetime
from distutils.version import StrictVersion
from jinja2 import Template
from semversioner import __version__

ROOTDIR = os.getcwd()
INITIAL_VERSION = '0.0.0'
DEFAULT_TEMPLATE = """# Changelog
Note: version releases in the 0.x.y range may introduce breaking changes.
{% for release in releases %}

## {{ release.id }}

{% for data in release.data %}
- {{ data.type }}: {{ data.description }}
{% endfor %}
{% endfor %}
"""


@click.group()
@click.option('--path', default=ROOTDIR, help="Base path. Default to current directory.")
@click.version_option(version=str(__version__.__version__))
@click.pass_context
def cli(ctx, path):
    semversioner_path_legacy = os.path.join(path, '.changes')
    semversioner_path_new = os.path.join(path, '.semversioner')
    semversioner_path = semversioner_path_new
    is_deprecated = False

    if os.path.isdir(semversioner_path_legacy) and not os.path.isdir(semversioner_path_new):
        is_deprecated = True
        semversioner_path = semversioner_path_legacy
    if not os.path.isdir(semversioner_path):
        os.makedirs(semversioner_path)

    dirname = os.path.join(semversioner_path, 'next-release')
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    ctx.obj = {
        'semversioner_path': semversioner_path,
        'is_deprecated': is_deprecated
    }


@cli.command('release', help="Release a new version.")
@click.pass_context
def cli_release(ctx):
    if ctx.obj['is_deprecated']:
        click.secho("WARN", bg='yellow', fg='black', nl=False)
        click.secho(" deprecated ", fg='magenta', nl=False)
        click.echo("Semversioner now uses '.semversioner' directory instead of '.changes'. Please, rename it to remove this message.")
    result = release(ctx.obj['semversioner_path'])
    click.echo(message="Successfully created new release: " + result['new_version'])

@cli.command('changelog', help="Print the changelog.")
@click.pass_context
def cli_changelog(ctx):
    changelog = generate_changelog(ctx.obj['semversioner_path'])
    click.echo(message=changelog, nl=False)


@cli.command('add-change', help="Create a new changeset file.")
@click.pass_context
@click.option('--type','-t', type=click.Choice(['major', 'minor', 'patch']), required=True)
@click.option('--description','-d', required=True)
def cli_add_change(ctx, type, description):
    result = add_change(ctx.obj['semversioner_path'], type, description)
    click.echo(message="Successfully created file " + result['path'])


@cli.command('current-version', help="Show the current version.")
@click.pass_context
def cli_current_version(ctx):
    version = get_current_version_number(ctx.obj['semversioner_path'])
    click.echo(message=version)


def add_change(semversioner_path, type, description):
    """ 
    Create a new changeset file.
  
    The method creates a new json file in the ``.semversioner/next-release/`` directory 
    with the type and description provided.
  
    Parameters
    -------
    semversioner_path (str): Semversioner path (e.g. /home/rgomis/.semversioner)
    type (str): Change type. Allowed values: major, minor, patch.
    description (str): Change description. 
  
    Returns
    -------
    path : str
        Absolute path of the file generated.
    """
    if not os.path.isdir(semversioner_path):
        os.makedirs(semversioner_path)
    next_release_path = os.path.join(semversioner_path, 'next-release')
    if not os.path.isdir(next_release_path):
        os.makedirs(next_release_path)
    
    parsed_values = {
        'type': type,
        'description': description,
    }

    change_type = parsed_values['type']
    filename = None
    while (filename is None or os.path.isfile(os.path.join(next_release_path, filename))):
        filename = '{type_name}-{datetime}.json'.format(
            type_name=parsed_values['type'],
            datetime="{:%Y%m%d%H%M%S}".format(datetime.datetime.utcnow()))

    with open(os.path.join(next_release_path, filename), 'w') as f:
        f.write(json.dumps(parsed_values, indent=2) + "\n")

    return { 
        'path': os.path.join(next_release_path, filename)
    }


def generate_changelog(semversioner_path):
    """ 
    Generates the changelog.
  
    The method generates the changelog based on the template file defined 
    in ``DEFAULT_TEMPLATE``.
  
    Parameters
    -------
    semversioner_path (str): Semversioner path (e.g. /home/rgomis/.semversioner)
  
    Returns
    -------
    str
        Changelog string.
    """
    releases = []
    for release_identifier in _sorted_releases(semversioner_path):
        with open(os.path.join(semversioner_path, release_identifier + '.json')) as f:
            data = json.load(f)
        data = sorted(data, key=lambda k: k['type'] + k['description'])
        releases.append({'id': release_identifier, 'data': data})
    return Template(DEFAULT_TEMPLATE, trim_blocks=True).render(releases=releases)


def release(semversioner_path):
    """ 
    Performs the release.
  
    The method performs the release by taking everything in ``next-release`` folder 
    and aggregating all together in a single JSON file for that release (e.g ``1.12.0.json``). 
    The JSON file generated is a list of all the individual JSON files from ``next-release``.
    After aggregating the files, it removes the ``next-release`` directory.
  
    Parameters
    -------
    semversioner_path (str): Semversioner path (e.g. /home/rgomis/.semversioner)
  
    Returns
    -------
    previous_version : str
        Previous version.
    new_version : str
        New version.
    """
    changes = []
    next_release_dir = os.path.join(semversioner_path, 'next-release')
    for filename in os.listdir(next_release_dir):
        full_path = os.path.join(next_release_dir, filename)
        with open(full_path) as f:
            changes.append(json.load(f))

    if len(changes) == 0:
        click.secho("Error: No changes to release. Skipping release process.", fg='red')
        sys.exit(-1)

    current_version_number = get_current_version_number(semversioner_path)
    next_version_number = _get_next_version_number(changes, current_version_number)
    click.echo("Releasing version: %s -> %s" % (current_version_number, next_version_number))
    
    release_json_filename = os.path.join(semversioner_path, '%s.json' % next_version_number)

    click.echo("Generated '" + release_json_filename + "' file.")
    with open(release_json_filename, 'w') as f:
        f.write(json.dumps(changes, indent=2, sort_keys=True))

    click.echo("Removing '" + next_release_dir + "' directory.")
    for filename in os.listdir(next_release_dir):
        full_path = os.path.join(next_release_dir, filename)
        os.remove(full_path)
    os.rmdir(next_release_dir)

    return {
        'previous_version': current_version_number,
        'new_version': next_version_number
    }


def get_current_version_number(semversioner_path):
    releases = _sorted_releases(semversioner_path)
    if len(releases) > 0:
        return releases[0]
    return INITIAL_VERSION


def _sorted_releases(semversioner_path):
    files = [f for f in os.listdir(semversioner_path) if os.path.isfile(os.path.join(semversioner_path, f))]
    releases=list(map(lambda x: x[:-len('.json')], files))
    releases=sorted(releases, key=StrictVersion, reverse=True)
    return releases


def _get_next_version_number(changes, current_version_number):
    release_type = sorted(list(map(lambda x: x['type'], changes)))[0]
    return increase_version(current_version_number, release_type)


def increase_version(current_version, release_type):
    """ 
    Returns a string like '1.0.0'.
    """
    # Convert to a list of ints: [1, 0, 0].
    version_parts = list(int(i) for i in current_version.split('.'))
    if release_type == 'patch':
        version_parts[2] += 1
    elif release_type == 'minor':
        version_parts[1] += 1
        version_parts[2] = 0
    elif release_type == 'major':
        version_parts[0] += 1
        version_parts[1] = 0
        version_parts[2] = 0
    return '.'.join(str(i) for i in version_parts)


def main():
    cli(obj={})


if __name__ == '__main__':
    main()
